/**
  *  \file flak/dlg-select.cc
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "flak/dlg-select.h"
#include "cpluslib/format.h"
#include "game/player.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "util/nls.h"
#include "ui/listbox.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/widgets.h"
#include "game/spec.h"
#include "ui/commands.h"

namespace {
    class FlakParticipantList : public UIListbox {
        // FIXME: const correctness
        FlakVcrEntry& entry;
        std::vector<int> content;
        enum {
            flShip   =  8192,
            flFleet  = 16384,
            flIdMask =  2047
        };
        void onMove(int);
     public:
        FlakParticipantList(FlakVcrEntry& entry);

        bool isAccessible(int pos);
        void drawPart(GfxCanvas& can, int from, int to);
        void getLayoutInfo(LayoutInfo& info);

        Signal1<int> sig_moved_to_ship;
    };

    class FlakShipInfo : public UIWidget {
        FlakVcrEntry& entry;
        int ship;
     public:
        FlakShipInfo(FlakVcrEntry& entry);

        void draw(GfxCanvas& can, UpdateInfo& ui);
        void getLayoutInfo(LayoutInfo& info);

        void setShip(int ship);
    };
};

FlakParticipantList::FlakParticipantList(FlakVcrEntry& entry)
    : UIListbox(0, 10, 20 * font_ems[FONT_NORMAL], 0), entry(entry)
{
    FlakBattle& b = entry.getBattle();
    for (int i = 0; i < b.getNumFleets(); ++i) {
        const FlakFleet& fleet = b.getFleetByNumber(i);
        content.push_back(i + flFleet);
        for (int s = 0; s < fleet.data.num_ships; ++s)
            content.push_back(s + fleet.data.first_ship + flShip);
    }
    setSize(content.size());
    scrollTo(0, 1);
    sig_moved.add(this, &FlakParticipantList::onMove);
}

void
FlakParticipantList::onMove(int)
{
    if (isAccessible(getCurrent()))
        sig_moved_to_ship.raise(content[getCurrent()] & flIdMask);
}

bool
FlakParticipantList::isAccessible(int pos)
{
    return (pos >= 0) && (pos < content.size()) && ((content[pos] & flShip) != 0);
}

void
FlakParticipantList::drawPart(GfxCanvas& can, int from, int to)
{
    FlakBattle& b = entry.getBattle();
    for (int i = from; i < to; ++i) {
        int ele = getTop() + i;
        if (ele < content.size()) {
            int id = content[ele] & flIdMask;
            if (content[ele] & flShip)
                drawEntry(can, UIColor::tc_Static, i,
                          "  " + getBasicString(b.getShipByNumber(id).data.name));
            else
                drawEntry(can, UIColor::tc_Red, i,
                          format(_(" Fleet %d (%s)"), id,
                                 getAdjName(b.getFleetByNumber(id).data.player)));
        } else {
            drawEntry(can, UIColor::tc_Faded, i, "");
        }
    }
}

void
FlakParticipantList::getLayoutInfo(LayoutInfo& info)
{
    info.flags     = li_Flexible;
    info.pref_size = GfxPoint(font_heights[FONT_NORMAL] * 10,
                            font_ems[FONT_NORMAL] * 20);
    info.min_size = GfxPoint(font_heights[FONT_NORMAL] * 3,
                           font_ems[FONT_NORMAL] * 5);
}



FlakShipInfo::FlakShipInfo(FlakVcrEntry& entry)
    : entry(entry), ship(0)
{ }

void
FlakShipInfo::draw(GfxCanvas& can, UpdateInfo& ui)
{
    int y = getExtent().y, dy = font_heights[FONT_NORMAL], x = getExtent().x, wi = getExtent().w;
    UISkin& sk = getSkin();

    FlakShip& sh = entry.getBattle().getShipByNumber(ship);

    sk.startDraw();
    sk.drawBackground(can, getExtent());
    sk.setColor(UIColor::tc_Static);

    sk.outTextF(can, FONT_NORMAL, x, y, wi, getBasicString(sh.data.name));
    y += dy;
    if (sh.isPlanet())
        sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("%s planet"), getAdjName(sh.data.player)));
    else
        sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("%s %s"), getAdjName(sh.data.player),
                                                       getHull(sh.data.hull).getShortName()));
    y += dy + dy/2;

    sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("Id: %d"), sh.data.id));
    y += dy;
    sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("Level: %d"), sh.data.level));
    y += dy;
    sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("Rating: %d/%d"), sh.data.rating, sh.data.compensation_rating));
    y += dy + dy/2;

    sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("Damage: %d%%"), sh.data.damage_init));
    y += dy;
    sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("Shield: %d%%, %d kt"), sh.data.shield_init, sh.data.mass));
    y += dy;
    if (!sh.isPlanet()) {
        sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("Crew: %d"), sh.data.crew_init));
        y += dy;
    }
    y += dy/2;

    if (sh.data.beam_count) {
        sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("%d x %s"), sh.data.beam_count,
                                                       getBeam(sh.data.beam_type).getName()));
        y += dy;
    }
    if (sh.data.torp_lcount) {
        sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("%d x %s launcher"),
                                                       sh.data.torp_lcount,
                                                       getTorp(sh.data.torp_type).getName()));
        y += dy;
        sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("  %d torpedo%!1{es%}"),
                                                       sh.data.torp_count));
        y += dy;
    }
    if (sh.data.bay_count) {
        sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("%d x fighter bay"), sh.data.bay_count));
        y += dy;
        sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("  %d fighter%!1{s%}"),
                                                       sh.data.fighter_count));
        y += dy;
    }
    if (sh.data.beam_count || sh.data.torp_lcount || sh.data.bay_count)
        y += dy/2;

    if (sh.data.ending_status == 0) {
        sk.outTextF(can, FONT_NORMAL, x, y, wi, _("Unit survived."));
    } else if (sh.data.ending_status == -1) {
        sk.outTextF(can, FONT_NORMAL, x, y, wi, _("Unit got destroyed."));
    } else if (sh.data.ending_status == sh.data.player) {
        sk.outTextF(can, FONT_NORMAL, x, y, wi, _("Unit got captured back."));
    } else {
        sk.outTextF(can, FONT_NORMAL, x, y, wi, format(_("Unit got captured by player %d."),
                                                       sh.data.ending_status));
    }
    y += dy;
    ui.update(getExtent());
}

void
FlakShipInfo::getLayoutInfo(LayoutInfo& info)
{
    /*
      NAME OF SHIP
      Owner, Type

      Id:     ID
      Level:  NAME
      Rating: RA/COM

      Damage: 100%
      Shield: 100%, 999kt
      Crew:   ...

      N x Foo Beam
      N x Foo Torp, N torps
      N fighter bays, N fighters

      Result: xxx
      */

    info.flags                     = 0;
    info.pref_size = info.min_size = GfxPoint(30 * font_ems[FONT_NORMAL],
                                            15 * font_heights[FONT_NORMAL]);
}

void
FlakShipInfo::setShip(int ship)
{
    if (ship != this->ship) {
        this->ship = ship;
        drawWidget();
    }
}

/**************************** FlakVcrSelector ****************************/

FlakVcrSelector::FlakVcrSelector(GFlakVcrDatabase& db)
    : db(db), current(0)
{ }

FlakVcrSelector::~FlakVcrSelector()
{ }

void
FlakVcrSelector::draw(GfxCanvas& can, UpdateInfo& ui)
{
    // FIXME: this is too simple
    UISkin& s = getSkin();
    FlakVcrEntry& b = db.getBattle(current);

    string_t text = format(_("Battle %d of %d: %d ships in %d groups\n"
                             "  Location: (%d,%d)\n"))
                                 << (current+1)
                                 << db.getCount()
                                 << b.getNumShips()
                                 << b.getNumFleets()
                                 << b.getPositionX()
                                 << b.getPositionY();
    int n = getExtent().h / font_heights[FONT_NORMAL] - 2;
    if (n > b.getNumShips())
        n = b.getNumShips();
    for (int i = 0; i < n; ++i) {
        FlakShip& sh = b.getBattle().getShipByNumber(i);
        if (sh.isPlanet())
            text += format(_("%s planet %d: %s\n"))
                << getAdjName(sh.data.player)
                << sh.data.id
                << getBasicStringN(sh.data.name, 20);
        else
            text += format(_("%s ship %d: %s\n"))
                << getAdjName(sh.data.player)
                << sh.data.id
                << getBasicStringN(sh.data.name, 20);
    }

    s.startDraw();
    s.drawBackground(can, getExtent());
    s.setSolid(false);
    s.setColor(UIColor::tc_Static);
    s.getParameters().font = fonts[FONT_NORMAL];
    outTextFormattedF(can, s.getParameters(), getExtent().x, getExtent().y,
                      getExtent().w, getExtent().h / font_heights[FONT_NORMAL],
                      text);
    ui.update(getExtent());
}

void
FlakVcrSelector::getLayoutInfo(LayoutInfo& info)
{
    // FIXME: this is too simple...
    info.flags = 0;
    info.min_size = info.pref_size =
        GfxPoint(30 * font_ems[FONT_NORMAL], 10 * font_heights[FONT_NORMAL]);
}

void
FlakVcrSelector::next(int)
{
    setCurrent(getCurrent()+1);
}

void
FlakVcrSelector::previous(int)
{
    setCurrent(getCurrent()-1);
}

void
FlakVcrSelector::showInfo()
{
    ObjectHolder h;

    static const UIKeyTranslator::Init xl[] = {
        SDLK_ESCAPE, cm_Escape,
        ' ', cm_Escape
    };

    UIWindow win(_("Combat Details"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    UIGroup content(UIHBoxLayout::instance5),
        buttons(UIHBoxLayout::instance5);
    UIFrameGroup listframe(UIHBoxLayout::instance0);

    FlakParticipantList fpl(db.getBattle(current));
    FlakShipInfo fsi(db.getBattle(current));
    fpl.sig_moved_to_ship.add(&fsi, &FlakShipInfo::setShip);
    
    win.setLayout(UIVBoxLayout::instance5);
    win.add(content);
    win.add(buttons);

    buttons.add(h.add(new UISpacer()));
    buttons.add(h.add(new UIButton(cm_Escape, _("Close"), SDLK_RETURN, UIButton::aStop)));
    buttons.add(h.add(new UIKeyTranslator(xl, countof(xl))));

    listframe.add(fpl);
    content.add(listframe);
    content.add(fsi);

    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    ui_root->run();
    ui_root->close(win);
}

void
FlakVcrSelector::setCurrent(int n)
{
    if (n >= db.getCount())
        n = db.getCount()-1;
    if (n < 0)
        n = 0;

    if (n != current) {
        current = n;
        sig_scrolled.raise();
        drawWidget();
    }
}

