/**
  *  \file flak/vis-flat.cc
  *  \brief FLAK Visualisation: Flat
  *
  *  This is the "scientific" fleet-based from-top flat view.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <cmath>
#include <iostream>
#include "flak/vis.h"
#include "gfx/palette.h"
#include "gfx/gfxcompl.h"
#include "flak/vis-sprites.h"

namespace {
    enum {
        SHIP_RADIUS    = 5
    };

    /// Sprite: fleet, flat view.
    class FlakFlatFleetSprite : public GfxSprite3D {
        FlakFlatVisualizer& vis;
        FlakBattle::fleet_t fleet;

        FlakFlatFleetSprite(const GfxRect& rect,
                            FlakFlatVisualizer& vis,
                            FlakBattle::fleet_t fleet)
            : GfxSprite3D(rect, 0, vis.layer_Ships), vis(vis), fleet(fleet)
            { }
        ~FlakFlatFleetSprite()
            { }
     public:
        void draw(GfxCanvas& can);
        void tick();
        static FlakFlatFleetSprite* create(FlakFlatVisualizer& vis, FlakBattle::fleet_t);
    };

    /// Sprite: background grid.
    class FlakFlatBackgroundSprite : public GfxSprite3D {
        FlakFlatVisualizer& vis;
     public:
        FlakFlatBackgroundSprite(FlakFlatVisualizer& vis);
        void draw(GfxCanvas& can);
    };
}

/************************** FlakFlatFleetSprite **************************/

/** Create fleet sprite.
    \param vis visualizer object
    \param f fleet number
    \return the fleet sprite, newly-allocated. Caller takes responsibility. */
FlakFlatFleetSprite*
FlakFlatFleetSprite::create(FlakFlatVisualizer& vis, FlakBattle::fleet_t f)
{
    FlakScreenPos p = vis.getScreenPosition(vis.getBattle().getFleetByNumber(f).status.position);
    return new FlakFlatFleetSprite(GfxRect(p.x - SHIP_RADIUS, p.y - SHIP_RADIUS,
                                           2*SHIP_RADIUS+1, 2*SHIP_RADIUS+1),
                                   vis, f);
}

void
FlakFlatFleetSprite::draw(GfxCanvas& can)
{
    FlakFleet& fl = vis.getBattle().getFleetByNumber(fleet);
    if (fl.isAlive()) {
        GfxParameters parm;
        parm.color = standard_colors[COLOR_WHITE];
        drawCircle(can, parm, getPosCX(), getPosCY(), SHIP_RADIUS);
    }
}

void
FlakFlatFleetSprite::tick()
{
    FlakFleet& f = vis.getBattle().getFleetByNumber(fleet);
    if (f.isAlive()) {
        setCenterPos2D(vis.getScreenPosition(f.status.position));
    } else {
        deleteThis();
    }
}

/************************ FlakFlatBackgroundSprite ***********************/

/*! \class FlakFlatBackgroundSprite
    This is the distance grid for the FlakFlatVisualizer. */
FlakFlatBackgroundSprite::FlakFlatBackgroundSprite(FlakFlatVisualizer& vis)
    : GfxSprite3D(vis.getManager().getBBox(), FlakFlatVisualizer::layer_Scene), vis(vis)
{ }

void
FlakFlatBackgroundSprite::draw(GfxCanvas& can)
{
    int min_x = vis.getCenterX() - vis.getRangeX()/2;
    int max_x = vis.getCenterX() + vis.getRangeX()/2;
    int min_y = vis.getCenterY() - vis.getRangeY()/2;
    int max_y = vis.getCenterY() + vis.getRangeY()/2;

    GfxParameters parm;
    parm.color = standard_colors[COLOR_DARK];
    parm.line_style = 0xAA;
    for (int x = min_x/10000; x <= max_x/10000; ++x) {
        GfxPoint p1 = vis.getScreenPosition(FlakPos(10000*x, min_y, 0));
        GfxPoint p2 = vis.getScreenPosition(FlakPos(10000*x, max_y, 0));
        drawLine(can, parm, p1.x, p1.y, p2.x, p2.y);
    }
    for (int y = min_y/10000; y <= max_y/10000; ++y) {
        GfxPoint p1 = vis.getScreenPosition(FlakPos(min_x, 10000*y, 0));
        GfxPoint p2 = vis.getScreenPosition(FlakPos(max_x, 10000*y, 0));
        drawLine(can, parm, p1.x, p1.y, p2.x, p2.y);
    }
}

/*************************** FlakFlatVisualizer **************************/

/*! \class FlakFlatVisualizer

    This is the standard flat visualizer, which shows exactly what
    happens within FLAK. It displays one circle per fleet (because
    fleets are the unit of movement in FLAK), as well as a distance
    grid every 10000 meters. */

/** Create new visualizer.
    \param battle battle to display
    \param area area of battle display */
FlakFlatVisualizer::FlakFlatVisualizer(FlakBattle& battle, GfxRect area)
    : FlakFancyVisualizer(battle, area), fleet_sprites(battle.getNumFleets()),
      ship_fleets(battle.getNumShips())
{
    // populate ship_fleets
    ship_t index = 0;
    for (fleet_t i = 1, max = getBattle().getNumFleets(); i < max; ++i) {
        ship_t here = getBattle().getFleetByNumber(i).data.first_ship;
        while (index < here)
            ship_fleets[index++] = i-1;
    }
    while (index < ship_fleets.size())
        ship_fleets[index++] = getBattle().getNumFleets()-1;

    // compute initial camera
    computeViewport();

    getManager().addSprite(new FlakFlatBackgroundSprite(*this));
}

FlakFlatVisualizer::~FlakFlatVisualizer()
{ }

void
FlakFlatVisualizer::everyTickAction()
{ }

void
FlakFlatVisualizer::computeViewport()
{
    center_x = 0;
    center_y = 0;

    int32 max_x = 19000, max_y = 19000;
    for (fleet_t i = 0, max = getBattle().getNumFleets(); i < max; ++i) {
        FlakPos p = getBattle().getFleetByNumber(i).status.position;
        max_x = std::max(max_x, int32(std::abs(p.x)));
        max_y = std::max(max_y, int32(std::abs(p.y)));
    }
    max_y = max_y * getManager().getBBox().w / getManager().getBBox().h;
    range = 2 * std::max(max_x, max_y) + 2000;
}

void
FlakFlatVisualizer::createSprites(bool is_startup)
{
    if (is_startup)
        computeViewport();

    // create fleets
    for (fleet_t i = 0, max = getBattle().getNumFleets(); i < max; ++i) {
        fleet_sprites[i] = FlakFlatFleetSprite::create(*this, i);
        manager.addSprite(fleet_sprites[i]);
    }

    // create objects
    for (int player = 1; player <= FLAK_NUM_OWNERS; ++player)
        for (FlakObject* o = getBattle().getPlayerObjects(player); o != 0; o = o->next)
            createObject(*o, FlakBattle::no_ship);
}

FlakScreenPos
FlakFlatVisualizer::getScreenPosition(FlakPos fp)
{
    const GfxRect& a = getManager().getBBox();
    return FlakScreenPos(a.x + (fp.x - center_x) * a.w / range + a.w / 2,
                         a.y - (fp.y - center_y) * a.w / range + a.h / 2,
                         0);
}

Ptr<GfxSprite3D>
FlakFlatVisualizer::getShipSprite(ship_t ship)
{
    return fleet_sprites[ship_fleets[ship]];
}

FlakBattle::ship_t
FlakFlatVisualizer::getShipAtPosition(GfxPoint pt)
{
    for (fleet_t i = 0, max = getBattle().getNumFleets(); i < max; ++i) {
        const FlakFleet& f = getBattle().getFleetByNumber(i);
        if (fleet_sprites[i] && fleet_sprites[i]->getPos().contains(pt)) {
            for (ship_t s = f.data.first_ship, smax = f.data.first_ship + f.data.num_ships; s < smax; ++s) {
                if (getBattle().getShipByNumber(s).isAlive())
                    return s;
            }
        }
    }
    return FlakBattle::no_ship;
}

FlakFancyVisualizer::EventResult
FlakFlatVisualizer::handleEvent(const UIEvent& evt)
{
    if (evt.type == ev_Key) {
        switch (evt.key.code) {
            // FIXME: key collision
         case ss_Shift + SDLK_LEFT:
            center_x -= range/15;
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case ss_Shift + SDLK_RIGHT:
            center_x += range/15;
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case ss_Shift + SDLK_UP:
            center_y += range/15;
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case ss_Shift + SDLK_DOWN:
            center_y -= range/15;
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case '+':
            range = 9*range/10;
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
         case '-':
            range = 10*range/9;
            getManager().tick();
            getManager().doUpdatesSilently();
            return eRedrawMe;
        }
    }
    return eNotHandled;
}

void
FlakFlatVisualizer::createObject(FlakObject& obj, ship_t source)
{
    if (source != FlakBattle::no_ship)
        markShipChanged(source);
    setObjectSprite(obj, FlakSimpleObjectSprite::create(*this, obj), true);
}
