/**
  *  \file game/act-fix.cc
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/action.h"
#include "game/objlist.h"
#include "game/global.h"
#include "game/ship.h"

/*********************** GStarbaseFixRecycleAction ***********************/

/** \class GStarbaseFixRecycleAction
    \brief Starbase shipyard order rule class.

    This class encapsulates all rules for a Fix/Recycle action. */

/** Initialize. */
GStarbaseFixRecycleAction::GStarbaseFixRecycleAction(GUniverse& univ, GPlanet& planet)
    : num_own(0), num_foreign(0), univ(univ), planet(planet)
{
    // FIXME: should follow actual host rules.
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        const GShip& s = univ.getShip(i);
        if (s.isExistant() && s.getPos() == planet.getPos()) {
            if (s.getOwner() == planet.getOwner())
                ++num_own;
            else
                ++num_foreign;
        }
    }
}

/** Check whether this starbase can do a "fix" command. */
bool
GStarbaseFixRecycleAction::canFix() const
{
    return num_own + num_foreign != 0;
}

/** Check whether this starbase can do a "recycle" command. */
bool
GStarbaseFixRecycleAction::canRecycle() const
{
    return num_own != 0;
}

/** Enumerate ships for a particular action.
    \param action desired action (sya_Fix, sya_Recycle).
    \param list   output list, should be empty. */
void
GStarbaseFixRecycleAction::enumerateShipsFor(int action, GObjectList& list)
{
    if (action == sya_None)
        return;
    ASSERT(action == sya_Recycle || action == sya_Fix);
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = univ.getShip(i);
        if (s.isExistant() && s.getPos() == planet.getPos())
            if (s.getOwner() == planet.getOwner() || action == sya_Fix)
                list.addObject(univ.getShipType(), s);
    }
}

bool
GStarbaseFixRecycleAction::setAction(int action, int ship_id)
{
    if (action == sya_None) {
        ship_id = 0;
    } else {
        if (ship_id <= 0 || ship_id >= NUM_SHIPS)
            return false;
        if (action != sya_Fix && action != sya_Recycle)
            return false;

        const GShip& ship = univ.getShip(ship_id);
        if (!ship.isExistant() || ship.getPos() != planet.getPos())
            return false;
        // FIXME: use actual host rule
        if (action != sya_Fix && ship.getOwner() != planet.getOwner())
            return false;
    }
    planet.setBaseShipyardOrder(action, ship_id);
    return true;
}
