/**
  *  \file game/msg-out.cc
  *  \brief Message Outbox Handling
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "cpluslib/assert.h"
#include "cpluslib/bitops.h"
#include "cpluslib/types.h"
#include "game/msg.h"
#include "game/player.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/recode.h"
#include "util/misc.h"

struct GOutbox::Message {
    string_t   text;
    GPlayerSet receivers;

    Message(const string_t text, GPlayerSet rec)
        : text(text), receivers(rec) { }
};

/** Header line for a universal message. */
const char GOutbox::UNIVERSAL_TEXT[] = "  <<< Universal Message >>>";

static const char CC_SELF_PREFIX[] = "<CC: ";
static const char CC_PREFIX[] = "CC: ";

/** Get "TO:" line for a receiver bitfield. */
static string_t
getReceiverText(GPlayerSet bits)
{
    /* Note: do not translate "Host" here, because this function is
       also used to generate title lines for sent messages */
    bits &= GPlayerSet::fromInteger(GOutbox::ALL_RECEIVERS_MASK);
    if (bits.empty()) {
        return _("Nobody");
    } else if (bits.isUnitSet()) {
        /* one receiver */
        for (int i = 1; i <= 12; ++i)
            if (bits.contains(i))
                return i == 12 ? string_t("Host") : getFullName(i);
        return "Huh?";
    } else {
        /* many receivers */
        string_t rv;
        for (int i = 1; i <= 12; ++i) {
            if (bits.contains(i)) {
                if (rv.length())
                    rv += " ";
                if (i == 12)
                    rv += "Host";
                else
                    rv += itoa(i);
            }
        }
        return rv;
    }
}

/******************************** GOutbox ********************************/

/*! \class GOutbox

    This class manages outgoing messages. In particular, it is
    responsible for converting a number of messages read from a
    message file into a uniform workable format, that is, it merges
    instances of a multi-message which have been separated by Maketurn
    or other file format requirements. */

/** Construct outbox.
    \param sender owner of this outbox: the player
                  whom all messages contained in this
                  messagebox belong to. */
GOutbox::GOutbox(int sender)
    : sender(sender), format(fWinplan)
{ }

GOutbox::~GOutbox()
{
    clear();
}

/** Discard all messages. */
void
GOutbox::clear()
{
    safeClear(messages);
    messages.clear();
}

int
GOutbox::getCount()
{
    return messages.size();
}

/** Get headers to use for a message to /receivers/.
    These headers are displayed, but not part of the stored text. */
string_t
GOutbox::getHeadersForDisplay(GPlayerSet receivers)
{
    receivers &= GPlayerSet::fromInteger(ALL_RECEIVERS_MASK);

    string_t recv_text = getReceiverText(receivers);
    string_t text = ::format(_("<<< Sub Space Message >>>\nFROM: %s\nTO: %s\n"),
                             getFullName(sender), recv_text);
    if (receivers.contains(GPlayerSet::fromInteger(UNIVERSAL_MASK))) {
        text += UNIVERSAL_TEXT;
        text += '\n';
    } else if (!receivers.isUnitSet()) {
        text += CC_PREFIX; // FIXME: i18n?
        text += recv_text;
        text += '\n';
    }
    return text;
}

/** Get text of message /index/ (0-based). The return value will
    include headers. */
string_t
GOutbox::getText(int index)
{
    ASSERT(index >= 0 && index < messages.size());

    return getHeadersForDisplay(messages[index]->receivers)
        + messages[index]->text;
}

string_t
GOutbox::getHeading(int index)
{
    ASSERT(index >= 0 && index < messages.size());
    GPlayerSet receivers = messages[index]->receivers;
    if (receivers.contains(GPlayerSet::fromInteger(UNIVERSAL_MASK)))
        return "Universal Message";
    else
        return ::format(_("To: %s"), getReceiverText(receivers));
}

/** Get raw text of a message. */
string_t
GOutbox::getRawText(int index)
{
    ASSERT(index >= 0 && index < messages.size());
    return messages[index]->text;
}

/** Get prefix for message when physically sent. The prefix can be
    carelessly concatenated to the message text.
    \param index    message number
    \param to       addressee of this incarnation of the message. */
string_t
GOutbox::getSendPrefix(int index, int to)
{
    ASSERT(index >= 0 && index < messages.size());
    GPlayerSet receivers = messages[index]->receivers & GPlayerSet::fromInteger(ALL_RECEIVERS_MASK);

    /* to 11 players, or 11 players + host */
    if (receivers.contains(GPlayerSet::fromInteger(UNIVERSAL_MASK)))
        return string_t(UNIVERSAL_TEXT) + '\n';

    receivers -= to;
    if (!receivers.empty()) {
        const char* pfx = (sender == to ? CC_SELF_PREFIX : CC_PREFIX);
        return pfx + getReceiverText(receivers) + '\n';
    }
    return string_t();
}

/** Get bitmask of receivers. 1<<1 means player 1, 1<<2 means player
    2, and so on up to 1<<12 for the host. */
GPlayerSet
GOutbox::getReceiverMask(int index)
{
    ASSERT(index >= 0 && index < messages.size());
    return messages[index]->receivers & GPlayerSet::fromInteger(ALL_RECEIVERS_MASK);
}

int
GOutbox::getStatus(int /*index*/)
{
    return 0;                   // FIXME?
}

void
GOutbox::setStatus(int /*index*/, int /*newstat*/)
{ }

/** Add message (send).
    \param text       message body without headers
    \param receivers  receiver set */
void
GOutbox::addMessage(string_t text, GPlayerSet receivers)
{
    safeAppend(messages, new Message(strRTrim(text), receivers));
}

static string_t
maybeStripHeaders(string_t msg, GPlayerSet rec)
{
    /* strip headers only from unicast messages */
    if (!rec.isUnitSet())
        return msg;

    const string_t first_line = strFirst(msg, "\n");
    if (first_line == GOutbox::UNIVERSAL_TEXT
        || first_line.compare(0, sizeof(CC_PREFIX)-1,      CC_PREFIX,      sizeof(CC_PREFIX)-1) == 0
        || first_line.compare(0, sizeof(CC_SELF_PREFIX)-1, CC_SELF_PREFIX, sizeof(CC_SELF_PREFIX)-1) == 0)
    {
        if (strRemove(msg, "\n"))
            return msg;
        else
            return string_t();  // silly people sending empty messages
    } else
        return msg;
}

/** Add message (from file). Unlike addMessage, this one attempts to
    merge multicast messages. */
void
GOutbox::addMessageFromFile(string_t text, GPlayerSet receivers)
{
    /* attempt to merge messages. Preconditions:
       - message box contains at least one message
       - receivers don't overlap
       - message bodies are identical, sans headers */
    text = strRTrim(text);
    string_t raw_text = maybeStripHeaders(text, receivers);

    if (!messages.empty()
        && (messages.back()->receivers & receivers).empty()
        && (maybeStripHeaders(messages.back()->text, messages.back()->receivers) == raw_text))
    {
        /* merge */
        messages.back()->receivers |= receivers;
        messages.back()->text = raw_text;
    } else {
        /* don't merge */
        addMessage(text, receivers);
    }
}

/** Delete all messages after the given index. */
void
GOutbox::deleteMessagesAfter(int index)
{
    if (index >= getCount())
        return;

    for (std::vector<Message*>::size_type m = index; m < messages.size(); ++m) {
        delete messages[m];
        messages[m] = 0;
    }
    messages.erase(messages.begin() + index, messages.end());
}

/** Delete message with the given index. All following messages are
    "shifted down" accordingly. */
void
GOutbox::deleteMessage(int index)
{
    ASSERT(index >= 0 && index < messages.size());

    delete messages[index];
    for (std::vector<Message*>::size_type n = index+1; n < messages.size(); ++n)
        messages[n-1] = messages[n];
    // FIXME: pedant alert
    messages.pop_back();
}

/** Encode message, in-place. Recodes the message for output, and
    encrypts it. */
void
encodeMessage(string_t& text)
{
    for (string_t::size_type i = 0; i < text.length(); ++i)
        if (text[i] == '\n')
            text[i] = 26;
        else
            text[i] = Recoder::getI2E().recodeCharacter(text[i]) + 13;
    
    /* terminate with a carriage return to avoid that
       people make fun of Akseli :-) */
    if (text.length() && text[text.length()-1] != 26)
        text += char(26);
}
