/**
  *  \file game/obj.h
  *  \brief Game Object Base Classes
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_OBJ_H
#define PCC_V2_GAME_OBJ_H

#include "game/coord.h"
#include "cpluslib/string.h"
#include "cpluslib/object.h"
#include "cpluslib/types.h"

/// Game object.
class GObject {
    bool is_played;
    bool is_marked;
    uint8 dirty_flags;

    GObject(const GObject&);
    GObject& operator=(const GObject&);
 public:
    /** Constructor. */
    GObject()
        : is_played(false), is_marked(false), dirty_flags(0)
        { }

    /** Destructor. */
    virtual ~GObject();

    string_t toString();

    /** Get name of this object.
        \return name, e.g. "USS Endeavour" */
    virtual string_t getName() const = 0;

    /** Get full name of this object.
        \return full name, e.g. "USS Endeavour (#42)" */
    virtual string_t getFullName() const = 0;

    /** Get Id number of this object. */
    virtual int getId() const = 0;

    /** Get owner of this object. */
    virtual int getOwner() const = 0;

    /** Check whether this object exists.
        \return true if this object exists, fals if not. */
    virtual bool isExistant() const = 0;

    /** Check whether this object is being played. */
    bool isPlayed() const
        { return is_played; }

    /** Set whether this object is being played. */
    void setIsPlayed(bool n)
        { is_played = n; }

    uint8 getDirtyFlags() const
        { return dirty_flags; }
    void setIsDirty(uint8 which)
        { dirty_flags |= which; }
    void setDirtyFlags(uint8 which)
        { dirty_flags = which; }
    bool isDirty(uint8 which) const
        { return (dirty_flags & which) != 0; }

    bool isMarked() const
        { return is_marked; }
    void setIsMarked(bool n)
        {
            setIsDirty(1);
            is_marked = n;
        }
};


/// Physical game object on map.
class GMapObject : public GObject {
 public:
    /** Constructor. */
    GMapObject()
        { }

    /** Destructor. */
    virtual ~GMapObject();

    /** Get position in game universe. */
    virtual GPoint getPos() const = 0;
};


/// Circular game object on map.
class GCircularObject : public GMapObject {
 public:
    virtual ~GCircularObject();

    /** Get object radius in light-years. */
    virtual int getRadius() const = 0;
};

#endif
