/**
  *  \file game/ship.cc
  *  \brief Ships
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include "game/ship.h"
#include "util/console.h"
#include "game/spec.h"
#include "util/nls.h"
#include "game/global.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "game/mission.h"
#include "game/config.h"

/******************************* Utilities *******************************/

static const TShipXY shipxy_init = { 0, 0, 0, 0 };

static int
computeMass(const TShip* ship)
{
    int mass = getHull(ship->hull).getMass()
        + ship->ammo
        + ship->ore_cargo[0] + ship->ore_cargo[1]
        + ship->ore_cargo[2] + ship->ore_cargo[3]
        + ship->colonists + ship->supplies;
    if (ship->torp_launcher_count)
        mass += ship->torp_launcher_count * getTorp(ship->torp_type).getMass();
    if (ship->beam_count)
        mass += ship->beam_count * getBeam(ship->beam_type).getMass();
    return mass;
}

/********************************* GShip *********************************/

/** Constructor. Creates a blank ship object for ship /id/. */
GShip::GShip(int id)
    : id(id), dat(0), target(0), shipxy(shipxy_init), orbit(0),
      src_flags(), target_flags(), rc_flag(0)
{ }

GShip::~GShip()
{
    delete dat;
    delete target;
}

/* Initialisation */

void
GShip::addData(const TShip& adat, GPlayerSet asrc_flags)
{
    if (!dat)
        dat = new TShip(adat);
    src_flags |= asrc_flags;
}

void
GShip::addTargetEntry(const TShipTarget& target, GPlayerSet flags)
{
    if (!this->target) {
        this->target = new TShipTarget(target);
    } else {
        // FIXME: Merge
    }
    target_flags |= flags;
}

void
GShip::addShipXYEntry(TShipXY axy)
{
    shipxy = axy;
}

void
GShip::addRCEntry(int flag)
{
    rc_flag = flag;
}

/* Checking */

void
GShip::internalCheck()
{
    if (dat) {
        /* we have a DAT entry. There must not be a target. It is not
           an error, though; allies receive our ships as targets. */
        if (target) {
            delete target;
            target = 0;
        }
        /* XY position must be correct */
        if (shipxy.x != dat->x || shipxy.y != dat->y || shipxy.owner != dat->owner) {
            console.write(LOG_WARN,
                          format(_("Ship #%d has inconsistent information -- correcting."), id));
            shipxy.x = dat->x;
            shipxy.y = dat->y;
            shipxy.owner = dat->owner;
        }
        shipxy.mass = computeMass(dat);
    } else if (target) {
        /* we have a target entry. */
//         if (dis) {
//             delete dis;
//             dis = 0;
//         }
        /* XY position must be correct */
        if (shipxy.x != target->x || shipxy.y != target->y || shipxy.owner != target->owner) {
            if (shipxy.owner)
                /* don't warn for FF targets. */
                console.write(LOG_WARN,
                              format(_("Ship #%d has inconsistent information -- correcting."), id));
            shipxy.x = target->x;
            shipxy.y = target->y;
            shipxy.owner = target->owner;
        }
    }
}

void
GShip::combinedCheck1(GUniverse& univ)
{
    if (!dat && isPlayed()) {
        /* we own it but can't see it? impossible! */
        delete target;
        target = 0;
        shipxy.x = shipxy.y = shipxy.owner = shipxy.mass = 0;
        console.write(LOG_WARN,
                      format(_("Ship #%d is ours but we don't see it -- deleted."), id));
        // FIXME: what if during combined check someone else sees that we play this ship?
        setIsPlayed(false);
        return;
    }

    /* This ship looks valid. Compute orbit. */
    orbit = univ.getPlanetAt(getPos());
    if (orbit)
        univ.getPlanet(orbit).setShipInOrbit(*this);
}

/* Accessors */

/** Get ship mass. Returns zero if unknown. */
int
GShip::getMass() const
{
    if (dat)
        return computeMass(dat);
    else
        return shipxy.mass;
}

/** Get hull Id. Returns zero if unknown. */
int
GShip::getHullId() const
{
    if (dat)
        return dat->hull;
    else if (target)
        return target->hull;
    else
        return 0;
}

/** Get real owner of ship. */
int
GShip::getRealOwner() const
{
    if (rc_flag > 0)
        return rc_flag;
    else
        return getOwner();
}

GPoint
GShip::getWaypoint() const
{
    return GPoint(shipxy.x + dat->waypoint_dx,
                  shipxy.y + dat->waypoint_dy);
}

bool
GShip::canDoSpecial(int hf) const
{
    return getHull().canDoSpecial(hf, getRealOwner(), true);
}

int
GShip::getWarp() const
{
    if (dat)
        return dat->warp;
    else if (target)
        return target->warp;
    else
        return -1;
}

bool
GShip::isHyperdriving() const
{
    if (!dat)
        return false;
    return canDoSpecial(hf_Hyperdrive)
        && dat->warp > 0
        && std::memcmp("HYP", dat->fcode, 3) == 0;
}

int
GShip::getFreeCargo() const
{
    return getHull().getCargo()
        - getCargo(el_Tritanium) - getCargo(el_Duranium) - getCargo(el_Molybdenum)
        - getAmmo() - getColonists() - getSupplies();
}

void
GShip::setName(const string_t& str)
{
    storeBasicString(dat->name, str);
    setIsDirty(ss_DirtyHead);
}

void
GShip::setFCode(const string_t& str)
{
    storeBasicString(dat->fcode, str);
    setIsDirty(ss_DirtyData);
}

void
GShip::setPE(int pe)
{
    if (dat->primary_enemy != pe) {
        dat->primary_enemy = pe;
        setIsDirty(ss_DirtyData);
    }
}

void
GShip::setWaypoint(GPoint pt)
{
    if (pt != getWaypoint()) {
        dat->waypoint_dx = pt.x - shipxy.x;
        dat->waypoint_dy = pt.y - shipxy.y;
        setIsDirty(ss_DirtyData);  // FIXME: DirtyXY?
    }
}

void
GShip::setWarp(int warp)
{
    if (warp != getWarp()) {
        dat->warp = warp;
        setIsDirty(ss_DirtyData);
    }
}

void
GShip::setMission(int m, int i, int t)
{
    if (m != dat->mission || i != dat->mission_intercept || t != dat->mission_tow) {
        dat->mission = m;
        dat->mission_intercept = i;
        dat->mission_tow = t;
        setIsDirty(ss_DirtyData);

        /* FIXME: I do not exactly like this, because it's slow, and
           it references getDisplayedTurn(). Can we do better? */
        if (const GMission* msn = GMissionList::getInstance().getMissionByNumber(m, GPlayerSet(raceId(getRealOwner())))) {
            if (msn->getFlags() & GMission::mf_Waypoint) {
                if (i > 0 && i <= NUM_SHIPS && getDisplayedTurn().getShip(i).isExistant()) {
                    setWaypoint(getDisplayedTurn().getShip(i).getPos());
                }
            }
        }
    }
}

/* GMapObject interface */

GPoint
GShip::getPos() const
{
    return GPoint(shipxy.x, shipxy.y);
}

string_t
GShip::getName() const
{
    if (dat)
        return getBasicString(dat->name);
    else if (target)
        return getBasicString(target->name);
    else
        return "<unknown>";
}

string_t
GShip::getFullName() const
{
    if (dat || target)
        return format(_("%s (Ship #%d)"), getName(), id);
    else
        return format(_("Ship %d"), id);
}

int
GShip::getId() const
{
    return id;
}

int
GShip::getOwner() const
{
    return shipxy.owner;
}

bool
GShip::isExistant() const
{
    return shipxy.owner != 0;
}

/* Cargo */

int
GShip::getAmmo() const
{
    GCargoType type = getBayCount() > 0 ? el_Fighters : GCargoType(el_Torps + getTorpType() - 1);
    return dat->ammo - cargo.getReserved(type);
}

/** Get cargo amount sans reservations. */
int
GShip::getCargo(GCargoType type) const
{
    return getCargoRaw(type) - cargo.getReserved(type);
}

/** Get cargo which is physically aboard the ship. */
int
GShip::getCargoRaw(GCargoType type) const
{
    switch (type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        return dat->ore_cargo[type];
     case el_Fighters:
        if (getBayCount() > 0)
            return dat->ammo;
        return 0;
     case el_Colonists:
        return dat->colonists;
     case el_Supplies:
        return dat->supplies;
     case el_Money:
        return dat->money;
     default:
        ASSERT(type >= el_Torps && type < el_COUNT);
        if (getTorpLauncherCount() > 0 && getTorpType() == (type - el_Torps + 1))
            return dat->ammo;
        return 0;
    }
}

/** Change amount that is on ship. */
void
GShip::changeCargoRaw(GCargoType type, int amount)
{
    if (!amount)
        return;

    setIsDirty(ss_DirtyData);
    switch (type) {
     case el_Neutronium:
     case el_Tritanium:
     case el_Duranium:
     case el_Molybdenum:
        dat->ore_cargo[type] += amount;
        break;
     case el_Fighters:
        dat->ammo += amount;
        break;
     case el_Colonists:
        dat->colonists += amount;
        break;
     case el_Supplies:
        dat->supplies += amount;
        break;
     case el_Money:
        dat->money += amount;
        break;
     default:
        ASSERT(type >= el_Torps && type < el_COUNT);
        dat->ammo += amount;
        break;
    }
}

/** Check whether ship is allowed to change waypoint. */
bool
GShip::getWaypointPermission() const
{
    // FIXME: real fleet implementation here
    return true;
}

/** Check whether ship wants to be cloned. This does not whether
    cloning actually is permitted (registeredness, hull restrictions,
    player restrictions). */
bool
GShip::isCloningAt(const GPlanet& planet) const
{
    return dat
        && planet.getPos() == getPos()
        && (dat->fcode[0] == 'c' && dat->fcode[1] == 'l' && dat->fcode[2] == 'n');
}

void
GShip::leaveFleet()
{
    // FIXME: real implementation here
}

bool
GShip::isTransporterActive(Transporter which_one) const throw()
{
    const TShipTransfer& t = getTransporter(which_one);
    return (t.ore[0] | t.ore[1] | t.ore[2] | t.ore[3] | t.colonists | t.supplies) != 0;
}

bool
GShip::isTransporterCancellable(Transporter which_one) const throw()
{
    // FIXME: this does not guarantee correctness in all cases: Take a
    // ship with 10 cargo room left, and 10 kt in the transporter. A
    // pending transaction adds 10 to the ship (i.e. it got the
    // guarantee that it'll be able to add this). When we now do
    // cancelTranporter() and commit that transaction, the ship is
    // overloaded.
    const TShipTransfer& t = getTransporter(which_one);
    long sum = (long) t.ore[0] + t.ore[1] + t.ore[2] + t.ore[3] + t.supplies + t.colonists;
    return sum == 0 || sum <= getFreeCargo();
}

void
GShip::cancelTranporter(Transporter which_one) throw()
{
    ASSERT(isTransporterCancellable(which_one));
    TShipTransfer& t = getTransporter(which_one);
    dat->ore_cargo[0] += t.ore[0];
    dat->ore_cargo[1] += t.ore[1];
    dat->ore_cargo[2] += t.ore[2];
    dat->ore_cargo[3] += t.ore[3];
    dat->colonists += t.colonists;
    dat->supplies += t.supplies;
    t.ore[0] = t.ore[1] = t.ore[2] = t.ore[3] = t.supplies = t.colonists = t.id = 0;

    setIsDirty(ss_DirtyData);
}

/******************************* GShipType *******************************/

GShipType::GShipType(GUniverse& univ)
    : GObjectType(NUM_SHIPS, univ)
{ }

bool
GShipType::isExistant(int index)
{
    if (index > 0 && index <= NUM_SHIPS) {
        GShip& s = getUniverse().getShip(index);
        if (s.isExistant() && s.isPlayed())
            return true;
    }
    return false;
}

GShip&
GShipType::getObject(int index)
{
    return getUniverse().getShip(index);
}

string_t
GShipType::getClassName()
{
    return string_t(_("Starship"));
}
