/**
  *  \file game/trn.h
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_TRN_H
#define PCC_V2_GAME_TRN_H

#include "game/struct.h"
#include "io/streams.h"
#include "game/reg.h"
#include "game/timestamp.h"

/** TRN Command codes. The names are the same as in UN-TRN, the file
    format list, and some utilities inspired by the above.
    Those marked "<-" have been renamed, which does not mean the list
    would be consistent now.*/
enum {
    /* Ship commands */
    tcm_ShipFIRST              = 1,
    tcm_ShipChangeFc           = 1,        // sid, 3 bytes FC
    tcm_ShipChangeSpeed        = 2,        // sid, 1 word
    tcm_ShipChangeWaypoint     = 3,        // sid, 2 words
    tcm_ShipChangeMission      = 4,        // sid, 1 word
    tcm_ShipChangePrimaryEnemy = 5,        // sid, 1 word
    tcm_ShipTowShip            = 6,        // sid, 1 word
    tcm_ShipChangeName         = 7,        // sid, 20 bytes
    tcm_ShipBeamDownCargo      = 8,        // sid, 7 words NTDMCS+id
    tcm_ShipTransferCargo      = 9,        // sid, 7 words NTDMCS+id
    tcm_ShipIntercept          = 10,       // sid, 1 word
    tcm_ShipChangeNeutronium   = 11,       // sid, 1 word
    tcm_ShipChangeTritanium    = 12,       // sid, 1 word
    tcm_ShipChangeDuranium     = 13,       // sid, 1 word
    tcm_ShipChangeMolybdenum   = 14,       // sid, 1 word
    tcm_ShipChangeSupplies     = 15,       // sid, 1 word
    tcm_ShipChangeColonists    = 16,       // sid, 1 word
    tcm_ShipChangeTorpedoes    = 17,       // sid, 1 word
    tcm_ShipChangeMoney        = 18,       // sid, 1 word
    tcm_ShipLAST               = 18,

    /* Planet commands */
    tcm_PlanetFIRST            = 21,
    tcm_PlanetChangeFc         = 21,       // pid, 3 bytes
    tcm_PlanetChangeMines      = 22,       // pid, 1 word <-
    tcm_PlanetChangeFactories  = 23,       // pid, 1 word
    tcm_PlanetChangeDefense    = 24,       // pid, 1 word
    tcm_PlanetChangeNeutronium = 25,       // pid, 1 dword
    tcm_PlanetChangeTritanium  = 26,       // pid, 1 dword
    tcm_PlanetChangeDuranium   = 27,       // pid, 1 dword
    tcm_PlanetChangeMolybdenum = 28,       // pid, 1 dword
    tcm_PlanetChangeColonists  = 29,       // pid, 1 dword
    tcm_PlanetChangeSupplies   = 30,       // pid, 1 dword
    tcm_PlanetChangeMoney      = 31,       // pid, 1 dword
    tcm_PlanetColonistTax      = 32,       // pid, 1 word
    tcm_PlanetNativeTax        = 33,       // pid, 1 word
    tcm_PlanetLAST             = 33,       // BuildBase is special
    tcm_PlanetBuildBase        = 34,       // pid, NO DATA

    /* Starbase commands */
    tcm_BaseFIRST              = 40,
    tcm_BaseChangeDefense      = 40,       // bid, 1 word
    tcm_BaseUpgradeEngineTech  = 41,       // bid, 1 word
    tcm_BaseUpgradeHullTech    = 42,       // bid, 1 word <-
    tcm_BaseUpgradeWeaponTech  = 43,       // bid, 1 word <-
    tcm_BaseBuildEngines       = 44,       // bid, 9 words
    tcm_BaseBuildHulls         = 45,       // bid, 20 words
    tcm_BaseBuildWeapons       = 46,       // bid, 10 words
    tcm_BaseBuildLaunchers     = 47,       // bid, 10 words
    tcm_BaseBuildTorpedoes     = 48,       // bid, 10 words
    tcm_BaseBuildFighters      = 49,       // bid, 1 word
    tcm_BaseFixRecycleShipId   = 50,       // bid, 1 word <-
    tcm_BaseFixRecycleShip     = 51,       // bid, 1 word action
    tcm_BaseChangeMission      = 52,       // bid, 1 word
    tcm_BaseBuildShip          = 53,       // bid, 7 words
    tcm_BaseUpgradeTorpTech    = 54,       // bid, 1 word
    tcm_BaseLAST               = 54,

    /* Rest */
    tcm_SendMessage            = 60,       // len, from, to, text
    tcm_ChangePassword         = 61,       // zero, 10 bytes
    tcm_SendBack               = 62        // recv, type, size, data
};

//! Turn File Handling.
class GTurnfile {
    struct CommandDef;
    static const CommandDef command_defs[];
    class CommandCompare;
    friend class CommandCompare;

    /* Turn file structure */
    TTurnHeader         trn_header;     ///< TRN header.
    TTaccomTurnHeader   taccom_header;  ///< Taccom header (TRN directory). Verbatim from turn file (offsets 1-based).
    TTurnDosTrailer     dos_trailer;    ///< DOS trailer.
    TTurnWindowsTrailer win_trailer;    ///< Windows trailer.
    Buffer<char>        data;           ///< Miscellaneous data. The TRN, usually ;)
    Buffer<int32>       offsets;        ///< Offsets of commands, pointing into data. zero-based. NOT the pointer array from the turn file!
    int                 version;        ///< TRN file sub-version (Winplan only).
    int                 features;       ///< TRN file features (bitfield trnf_Xxx).
    int                 trn_placement;  ///< Taccom: place TRN before Nth attachment.

    /* Internal stuff */
    bool dirty;                 ///< True if data is dirty. If false, data is a valid turn file.

    void updateTurnfile(Buffer<char>& data, Buffer<int32>& offsets);
    void parseTurnfile(int32 offset, int32 length);
    void parseTurnfileHeader(Stream& stream, int32 offset, int32 length);
    void checkRange(int32 offset, int32 length);
    void makeCommands(int id, int low, int up, const char* old, const char* neu);
 public:
    /** Feature Flags. */
    enum {
        trnf_Winplan = 1,       ///< File contains Winplan trailer.
        trnf_Taccom  = 2        ///< File contains Taccom-style attachments.
    };
    /** Manifest Constants. */
    enum {
        CURRENT_VERSION = 1,    ///< We know TRN sub-version 1 (Host 3.22).
        MAX_ATTACH      = 10    ///< Maximum number of attachments (Taccom).
    };
    /** Command Types. */
    enum CommandType {
        UndefinedCommand,       ///< Command not known to us.
        ShipCommand,            ///< Ship command. First word is Id.
        PlanetCommand,          ///< Planet command. First word is Id.
        BaseCommand,            ///< Base command. First word is Id.
        OtherCommand            ///< Other command. Password, Message, Sendback.
    };

    GTurnfile(int player, GTimestamp time);
    GTurnfile(Stream& stream, bool full_parse = true);
    ~GTurnfile();

    /** \name Accessors */
    //@{
    /** Get player number. */
    int getPlayer() const
        { return trn_header.player; }
    /** Get turn timestamp. */
    GTimestamp getTimestamp() const
        { return GTimestamp(trn_header.timestamp); }
    /** Get number of commands stored in this turn. This is not
        necessary the number of commands the turn will have when
        written to disk, since there might be some deleted or invalid
        ones. */
    int getNumCommands() const
        { return offsets.getSize(); }
    /** Get bitfield of features active in this turn file. */
    int getFeatures() const
        { return features; }
    /** Get sub-version of turn file. Only valid for Winplan turns.
        Returns the "xy" in "file format 3.5xy" (currently either 0 or
        1). */
    int getVersion() const
        { return version; }

    void setVersion(int n);
    void setTimestamp(const GTimestamp& time);

    int         getCommandCode(int index) const;
    int         getCommandLength(int index) const;
    int         getCommandId(int index) const;
    CommandType getCommandType(int index) const;
    int32       getCommandPosition(int index) const;
    const char* getCommandName(int index) const;
    const char* getCommandData(int index) const;
    //@}

    static CommandType getTypeForCode(int code); // FIXME: these names suck
    static const char* getNameForCode(int code);
    static int         getRecordIndexForCode(int code);

    void setFeatures(int f);

    /** \name Header Accessors */
    //@{
    /** Get Windows (v3.5) Trailer. */
    const TTurnWindowsTrailer& getWindowsTrailer() const
        { return win_trailer; }
    /** Get DOS (v3.0) Trailer. */
    const TTurnDosTrailer& getDosTrailer() const
        { return dos_trailer; }
    /** Get turn header. */
    const TTurnHeader& getTurnHeader() const
        { return trn_header; }
    //@}

    /* Modificators */
    void addCommand(int cmd, int id);
    void addCommand(int cmd, int id, const char* data, std::size_t len);
    void addData(const char* data, std::size_t len);

    void deleteCommand(int index);

    void update();
    void updateTrailer();
    int32 computeTurnChecksum() const;
    void setTemplock(const uint32* data);

    void write(Stream& stream);

    void setRegInfo(const GRegInfo& reg, int turn_nr);

    /* Iteration */
    void sortCommands();

    /* Maketurn */
    // FIXME: this is probably too much an indirection
    void makeShipCommands(int id, const char* old, const char* neu);
    void makePlanetCommands(int id, const char* old, const char* neu);
    void makeBaseCommands(int id, const char* old, const char* neu);

    /** \name Taccom Accessors. */
    //@{
    int  addFile(const char* file_data, std::size_t len, const string_t& name);
    int  addFile(const Buffer<char>& data, const string_t& name);
    void deleteFile(int index);
    int  getFileCount() const;
    /** Get raw TacCom header. Since that currently is only used by
        un-trn, I didn't bother to write accessors for everything. */
    const TTaccomTurnHeader& getTaccomHeader() const
        { return taccom_header; }
    /** Get position of turn data in TacCom container. */
    int getTaccomTurnPlace() const
        { return trn_placement; }
    //@}
};

#endif
