/**
  *  \file game/univ.cc
  */

#include <algorithm>
#include "game/univ.h"
#include "game/planet.h"
#include "cpluslib/misc.h"
#include "game/ship.h"
#include "game/ion.h"
#include "game/host.h"
#include "game/formula.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/mine.h"
#include "util/misc.h"

/** Create blank universe. */
GUniverse::GUniverse()
    : turn_nr(0)
{
    std::fill_n(planets, countof(planets), (GPlanet*) 0);
    std::fill_n(ships,   countof(ships),   (GShip*) 0);
    std::fill_n(storms,  countof(storms),  (GIonStorm*) 0);
    ship_type   = new GShipType(*this);
    planet_type = new GPlanetType(*this);
    base_type   = new GBaseType(*this);
    storm_type  = new GIonStormType(*this);
}

/** Destroy universe. */
GUniverse::~GUniverse()
{
    done();
    delete ship_type;
    delete planet_type;
    delete base_type;
    delete storm_type;
}

/** Initialize universe. This fills in all objects. */
void
GUniverse::init()
{
    for (int i = 0; i < NUM_PLANETS; ++i)
        planets[i] = new GPlanet(i+1);
    for (int i = 0; i < NUM_SHIPS; ++i)
        ships[i] = new GShip(i+1);
    for (int i = 0; i < NUM_STORMS; ++i)
        storms[i] = new GIonStorm(i+1);
    clearAllMines();
}

/** Uninitialize universe. */
void
GUniverse::done()
{
    for (int i = 0; i < NUM_PLANETS; ++i) {
        delete planets[i];
        planets[i] = 0;
    }
    for (int i = 0; i < NUM_SHIPS; ++i) {
        delete ships[i];
        ships[i] = 0;
    }
    for (int i = 0; i < NUM_STORMS; ++i) {
        delete storms[i];
        storms[i] = 0;
    }
    clearAllMines();
}

/** Postprocess universe. Call this to make sure structural changes
    propagate. */
void
GUniverse::postprocess(const GPlayerSet playing_set)
{
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        GPlanet& p = getPlanet(i);
        p.internalCheck();
        p.setIsPlayed(p.isExistant() && playing_set.contains(p.getOwner()));
    }
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = getShip(i);
        s.internalCheck();
        s.setIsPlayed(s.isExistant() && playing_set.contains(s.getOwner()));
    }
    for (MinefieldMap::iterator i = minefields.begin(); i != minefields.end(); /* empty */) {
        i->second->internalCheck(*this);
        if (i->second->isExistant()) {
            ++i;
        } else {
            delete i->second;
            minefields.erase(i++);
        }
    }
    for (int i = 1; i <= NUM_SHIPS; ++i)
        getShip(i).combinedCheck1(*this);
    for (int i = 1; i <= NUM_PLANETS; ++i)
        getPlanet(i).combinedCheck2(*this);

    ship_type->init();
    planet_type->init();
    base_type->init();
    storm_type->init();
}

/** Get planet at location, with warp wells. If there is no planet at
    the location, look whether the point is in the warp well of one.
    \param pt Location, need not be normalized
    \param gravity true to look into warp well. If false, function
                   behaves exactly like getPlanetAt(int).
    \return Id of planet, or zero if none */
int
GUniverse::getPlanetAt(const GPoint pt, bool gravity) const
{
    int rv = getPlanetAt(pt);
    if (!rv && gravity)
        rv = getGravityPlanetAt(pt);
    return rv;
}

/** Get planet at location.
    \param pt Location, need not be normalized
    \return Id of planet, or zero if none */
int
GUniverse::getPlanetAt(const GPoint pt) const
{
    const GPoint real_pos = pt.getNormalized();
    for (int i = 1; i <= NUM_PLANETS; ++i)
        if (getPlanet(i).isExistant() && getPlanet(i).getPos() == pt)
            return i;
    return 0;
}

/** Get planet from warp well location.
    \param pt Location
    \pre getPlanetAt(pt) == 0
    \return Id of planet if pt is in its warp wells, 0 otherwise */
int
GUniverse::getGravityPlanetAt(GPoint pt) const
{
    /* easy case */
    if (!config.AllowGravityWells)
        return 0;

    if (host.isPHost()) {
        /* PHost gravity wells */
        int sqs = square(config.GravityWellRange);
        for (int i = NUM_PLANETS; i > 0; --i) {
            const GPlanet& p = getPlanet(i);
            if (p.isExistant()) {
                if (config.RoundGravityWells) {
                    if (distanceSquared(p.getPos(), pt) <= sqs)
                        return i;
                } else {
                    GPoint p2 = p.getPos().getNearestTo(pt);
                    if (square(p2.x - pt.x) <= sqs && square(p2.y - pt.y) <= sqs)
                        return i;
                }
            }
        }
        return 0;
    } else {
        /* THost gravity wells: round, 3 ly, not wrapped, "cumulative" */
        int pid = 0;
        for (int i = 1; i <= NUM_PLANETS; ++i) {
            const GPlanet& p = getPlanet(i);
            if (p.isExistant() && distanceSquaredRaw(p.getPos(), pt) <= 9) {
                pt = p.getPos();  // (!)
                pid = i;
            }
        }
        return pid;
    }
}

/** Format name of a planet. */
static string_t
formatPlanetName(const GUniverse& univ, int pid, int flags)
{
    return format(((flags & locs_Orbit)
                   ? ((flags & locs_Verbose)
                      ? _("Orbit of %s (Planet #%d)")
                      : _("Orbit of %s (#%d)"))
                   : ((flags & locs_Verbose)
                      ? _("%s (Planet #%d)")
                      : _("%s (#%d)"))),
                  univ.getPlanet(pid).getName(), pid);
}

/** Get name of a location in human-readable form.
    \param pt     location
    \param flags  details of requested string
    - locs_Ships: also use ship names as location names, not just planets
    - locs_WW: return "near PLANETNAME" if applicable
    - locs_Orbit: say "Orbit of PLANET" if applicable
    - locs_Verbose: be more verbose */
string_t
GUniverse::getLocationName(GPoint pt, int flags) const
{
    /* is it a bird? */
    int pid = getPlanetAt(pt);
    if (pid)
        return formatPlanetName(*this, pid, flags);

    /* is it a plane? */
    if (flags & locs_Ships) {
        int sid = getAnyShipAt(pt);
        if (sid)
            return getShip(sid).getFullName();
    }
    if (flags & locs_WW) {
        pid = getGravityPlanetAt(pt);
        if (pid)
            return format(_("near %s"), formatPlanetName(*this, pid, flags & ~locs_Orbit));
    }
    return format((flags & locs_Verbose) ? _("Deep Space (%d,%d)") : "(%d,%d)",
                  pt.x, pt.y);
}

/** Get ship at position. Any race does. This is mainly used for naming
    locations, and for things like to tell whether a "L" command would succeed.
    \param pt position to check
    \returns Id number of a ship at position \c pt, or zero if none. */
int
GUniverse::getAnyShipAt(GPoint pt) const
{
    pt = pt.getNormalized();
    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (getShip(i).isExistant() && getShip(i).getPos() == pt)
            return i;
    return 0;
}

/** Count own ships at position.
    \param pt position to check
    \returns number of currently-played ships which are at the
    specified position \c pt. */
int
GUniverse::getNumberOfPlayedShipsAt(GPoint pt) const
{
    int n = 0;
    pt = pt.getNormalized();
    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (getShip(i).isExistant() && getShip(i).getPos() == pt && getShip(i).isPlayed())
            ++n;
    return n;
}

/** Delete all minefields. */
void
GUniverse::clearAllMines()
{
    for (MinefieldMap::iterator i = minefields.begin(); i != minefields.end(); ++i) {
        delete i->second;
        i->second = 0;
    }
    MinefieldMap().swap(minefields);
}

/** Get Minefield.
    \param id     minefield you want
    \param cause  cause for this request. mfa_Xxx or -1.
    \post cause >= 0 ==> return value != NULL */
GMinefield*
GUniverse::getMinefield(int id, int cause)
{
    MinefieldMap::iterator i = minefields.find(id);
    if (i == minefields.end()) {
        if (cause >= 0) {
            GMinefield* mf = new GMinefield(id, cause);
            safeInsert(minefields, id, mf);
            return mf;
        } else {
            return 0;
        }
    } else {
        if (cause > i->second->getAction())
            i->second->setAction(cause);
        return i->second;
    }
}

/** Perform all updates. This will poll all updatable objects, and
    raise the appropriate signals.
    - planets: sig_planet_changed
    - ships: sig_ship_changed */
void
GUniverse::doUpdates()
{
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        GPlanet& p = getPlanet(i);
        if (p.getDirtyFlags()) {
            /* call sig_planet_changed in a loop, in case it causes
               the planet to be modified further. Make sure that we
               don't run into an endless loop, though. */
            int flags = 0;
            do {
                flags |= p.getDirtyFlags();
                sig_planet_changed.raise(i);
            } while (p.getDirtyFlags() & ~flags);
            p.setDirtyFlags(0);
        }
    }

    for (int i = 1; i <= NUM_SHIPS; ++i) {
        GShip& s = getShip(i);
        if (s.getDirtyFlags()) {
            int flags = 0;
            do {
                flags |= s.getDirtyFlags();
                sig_ship_changed.raise(i);
            } while (s.getDirtyFlags() & ~flags);
            s.setDirtyFlags(0);
        }
    }
}
