/**
  *  \file util/idalloc.h
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_UTIL_IDALLOC_H
#define PCC_V2_UTIL_IDALLOC_H

#include <vector>
#include "cpluslib/types.h"

/** Id Number Allocator.

    This class provides a means of allocating Id numbers associated
    with more complicated (but POD) objects.

    Id numbers are non-negative integers, but the Id 0 is permitted. 
    You add an object to this allocator by calling addObject(), this
    returns an Id you can use to retrieve this object later. You
    cannot affect the Id numbers you get. You can store an object
    multiple times. The object stored must be POD. */
template<typename POD, typename ID = uint32>
class IdAllocator {
 public:
    /** Value type (values contained in this container). */
    typedef POD value_type;
    /** Id type (Id numbers have this type). */
    typedef ID  id_type;
    /** Null Id. You can use this value to mean "no Id number". */
    static const id_type nil = id_type(-1);
 private:
    union Elem {
        value_type value;
        id_type    next;
    };

    std::vector<Elem> elements;
    id_type           first_free;
    
 public:
    IdAllocator();
    ~IdAllocator();

    /** Add object.
        \return associated Id
        \inv getObject(addObject(x)) == x */
    id_type addObject(value_type value);
    /** Get object by Id.
        \return associated object.
        \pre the Id has been obtained from getObject() and not yet passed to
        removeObject(). */
    value_type getObject(id_type id) const
        { return elements[id].value; }
    /** Remove an object. This removes the association between object
        and Id; the Id will be re-used. */
    void removeObject(id_type id)
        {
            elements[id].next = first_free;
            first_free = id;
        }
    /** Remove all associations. */
    void clear();
};

template<typename POD, typename ID>
IdAllocator<POD,ID>::IdAllocator()
    : first_free(nil)
{ }

template<typename POD, typename ID>
IdAllocator<POD,ID>::~IdAllocator()
{ }

template<typename POD, typename ID>
ID
IdAllocator<POD,ID>::addObject(value_type value)
{
    id_type rv;
    if (first_free == nil) {
        rv = elements.size();
        elements.push_back(Elem());
    } else {
        rv = first_free;
        first_free = elements[first_free].next;
    }
    elements[rv].value = value;
    return rv;
}

template<typename POD, typename ID>
void
IdAllocator<POD,ID>::clear()
{
    std::vector<Elem>().swap(elements);
    first_free = nil;
}

#endif
