/**
  *  \file util/nls.cc
  *  \brief National Language Support
  *
  *  This module provides message translation in a gettext-like fashion.
  *  We don't use GNU gettext because it's GPL and because its data files
  *  are not portable between architectures. Otherwise, this could be
  *  easily be changed to use gettext when available.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <algorithm>
#include "arch/platform.h"
#include "cpluslib/buffer.h"
#include "cpluslib/types.h"
#include "io/dirs.h"
#include "io/io.h"
#include "io/streams.h"
#include "util/bsearch.h"
#include "util/console.h"
#include "util/except.h"
#include "util/nls.h"
#include "cpluslib/format.h"
#include "util/recode.h"
#include "cpluslib/assert.h"
#include "util/misc.h"

/* File format is:
   +0  8 BYTEs   magic: "CClang0",26
   +8    DWORD   number of entries
  +12    DWORD   pointer to "input pointer" array
  +16    DWORD   pointer to "output pointer" array
  +20    DWORD   pointer to input data
  +24    DWORD   length of input data
  +28    DWORD   pointer to output data
  +32    DWORD   length of output data */

// static StringTranslator main_translator;

static StringTranslator* main_translator;

const char*
translate(const char* in)
{
    return main_translator ? main_translator->translate(in) : in;
}

StringTranslator*
loadLanguageDatabase(const char* domain, const char* fn)
{
    string_t res     = Platform::getResourceDirectory();
    string_t fname;
    if (*domain) {
        fname += domain;
        fname += "_";
    }
    fname += fn;

    StringTranslator* rv;
    if (!StringTranslator::getCreateTranslation(fname, rv)) {
        string_t full_fn = Directory::makeFileName(res, fname + ".lang");
        Ptr<Stream> s = Stream::openFileNT(full_fn, Stream::C_READ);
        if (s)
            rv->loadLanguageFile(*s);
    }
    return rv;
}

void
initNLS()
{
    try {
        string_t code = Platform::getLanguageCode();
        string_t::size_type n = code.find_first_of("_:-/.\\");
        if (n != code.npos)
            code.erase(n);
        if (!code.length())
            code = "en";
        main_translator = loadLanguageDatabase("", code.c_str());
    }
    catch(FileOpenException& e) {
        /* ignore */
    }
    catch(std::exception& e) {
        console.write(LOG_ERROR, string_t("Unable to initialize language support: ") + e.what());
    }
}

/******************************* StringPool ******************************/

/*! \class StringPool
    \brief Pool of Bytes (for C strings)

    This class provides a simple allocator. It allocates large blocks
    of memory from the system allocator, and hands out smaller chunks
    of that.

    StringPool is intended to be used for C strings, although nothing
    in this allocator assumes anything about strings. However, since
    no form of alignment is maintained, its use for different data
    structures is limited.

    Freeing individual entries is not supported. */

const std::size_t StringPool::MIN_POOL;

/** Constructor. Create new empty pool. */
StringPool::StringPool()
    : current_pool(0), current_free(0)
{ }

/** Destructor. */
StringPool::~StringPool()
{
    clear();
}

/** Discard all contents. */
void
StringPool::clear()
{
    for (std::vector<Pool>::iterator i = pools.begin(); i != pools.end(); ++i)
        delete[] i->ptr;
    std::vector<Pool>().swap(pools);
    current_free = 0;
    current_pool = 0;
}

/** Allocate given size.
    \param size amount to allocate
    \return pointer to \c size bytes of free memory */
char*
StringPool::allocate(std::size_t size)
{
    if (current_free < size) {
        // make room for one element
        if (pools.empty() || pools.back().ptr != 0)
            pools.push_back(Pool());

        Pool new_pool;
        new_pool.size = size;
        if (new_pool.size < MIN_POOL)
            new_pool.size = MIN_POOL;

        new_pool.ptr = new char[new_pool.size];
        pools.back() = new_pool;

        if (new_pool.size == size)
            return new_pool.ptr;

        current_pool = new_pool.ptr;
        current_free = new_pool.size;
    }

    ASSERT(current_free >= size);
    current_free -= size;
    return current_pool + current_free;
}

/** Recode all content. Applies recoder \c r to all bytes in this
    pool.
    \param r recoder to apply */
void
StringPool::recode(const Recoder& r)
{
    for (std::vector<Pool>::iterator i = pools.begin(); i != pools.end(); ++i) {
        if (i->ptr) {
            // FIXME: this also recodes the uninitialized
            // portion at the end of each chunk
            r.recodeArray(i->ptr, i->size);
        }
    }
}

/**************************** StringTranslator ***************************/

const int StringTranslator::MAX_NUMBERED_ENTRY;

StringTranslator::StringTranslator()
{ }

StringTranslator::~StringTranslator()
{ }

bool
StringTranslator::Compare::operator()(const char* l, const char* r) const
{
    using std::strcmp;
    return strcmp(l, r) < 0;
}

const char*
StringTranslator::translate(const char* in)
{
    std::map<const char*, const char*, Compare>::const_iterator i = data.find(in);
    if (i != data.end()) {
        ensureCorrectOutputCharset();
        return i->second;
    } else {
        return in;
    }
}

const char*
StringTranslator::getNumberedEntry(int n)
{
    if (n >= 0 && unsigned(n) < numbered_data.size()) {
        ensureCorrectOutputCharset();
        return numbered_data[n];
    } else {
        return 0;
    }
}

void
StringTranslator::ensureCorrectOutputCharset()
{
    CharacterSet new_output_charset = Recoder::getInternCS();
    if (current_output_charset != new_output_charset) {
        output_strings.recode(Recoder(current_output_charset, new_output_charset));
        current_output_charset = new_output_charset;
    }
}

void
StringTranslator::addTranslation(const char* ori, const char* translated)
{
    using namespace std;

    if (strcmp(ori, translated) == 0)
        return;

    size_t olen = strlen(ori), tlen = strlen(translated);
    char* optr = input_strings.allocate(olen + 1);
    char* tptr = output_strings.allocate(tlen + 1);
    memcpy(optr, ori, olen + 1);
    memcpy(tptr, translated, tlen + 1);
    data[optr] = tptr;
}

bool
StringTranslator::addNumberedEntry(int n, const char* text)
{
    using namespace std;

    if (n < 0 || n > MAX_NUMBERED_ENTRY)
        return false;

    if (numbered_data.size() < unsigned(n) + 1)
        numbered_data.resize(n + 1);

    size_t tlen = strlen(text);
    char* tptr = output_strings.allocate(tlen + 1);
    memcpy(tptr, text, tlen + 1);
    numbered_data[n] = tptr;
    return true;
}

void
StringTranslator::loadLanguageFile(Stream& s)
{
    using namespace std;

    enum {
        i_Count,
        i_InPtr,
        i_OutPtr,
        i_InText,
        i_InSize,
        i_OutText,
        i_OutSize
    };

    /* start with a simple, inefficient implementation: we read
       everything into a buffer and process that, instead of reading
       directly into the StringPools */
    Buffer<char> language_db;
    s.readBuffer(language_db);
    if (language_db.getSize() < 36 || memcmp(language_db.ptr(), "CClang0\032", 8) != 0)
        throw FileFormatException(format("File `%s' is not a language file", s.getName()));

    uint32 data[7]; // 0-count 1-inptr 2-outptr 3-intxt 4-insize 5-outtxt 6-outsize
    getUint32(&language_db[8], data, 7);
    if (   data[i_InPtr]   >= language_db.getSize()
        || data[i_OutPtr]  >= language_db.getSize()
        || data[i_InText]  >= language_db.getSize()
        || data[i_OutText] >= language_db.getSize()
        || 8*data[i_Count] > language_db.getSize() - data[i_InPtr]
        || 8*data[i_Count] > language_db.getSize() - data[i_OutPtr]
        || data[i_InSize]  > language_db.getSize() - data[i_InText]
        || data[i_OutSize] > language_db.getSize() - data[i_OutText])
    {
        throw FileFormatException(format("File `%s' has an invalid format", s.getName()));
    }

    /* File is valid */
    for (uint32 i = 0; i < data[i_Count]; ++i) {
        uint32 in_ptr  = getUint32(language_db, data[i_InPtr] + 8*i);
        uint32 in_len  = getUint32(language_db, data[i_InPtr] + 8*i + 4);
        uint32 out_ptr = getUint32(language_db, data[i_OutPtr] + 8*i);
        uint32 out_len = getUint32(language_db, data[i_OutPtr] + 8*i + 4);
        if (in_ptr > data[i_InSize] || out_ptr > data[i_OutSize]
            || in_len > data[i_InSize] - in_ptr || out_len > data[i_OutSize] - out_ptr)
        {
            /* FIXME: check zero-terminatedness */
            throw FileFormatException(format("File `%s' has an invalid format", s.getName()));
        }
        addTranslation(&language_db[data[i_InText] + in_ptr],
                       &language_db[data[i_OutText] + out_ptr]);
    }

    /* character set hacking */
    string_t n = translate("");
    current_output_charset = 0;
    do {
        string_t line = strFirst(n, "\n");
        if (strCaseCmp(line.substr(0, 14), "X-PCC-Charset:") == 0) {
            current_output_charset = Recoder::getCharsetByName(strTrim(line.substr(14)));
            break;
        }
    } while (strRemove(n, "\n"));
}

void
StringTranslator::clear()
{
    std::vector<char*>().swap(numbered_data);
    std::map<const char*,const char*, Compare>().swap(data);
    current_output_charset = 0;
    input_strings.clear();
    output_strings.clear();
}

static std::map<string_t,StringTranslator*> translations;

StringTranslator*
StringTranslator::getTranslation(const string_t name)
{
    std::map<string_t,StringTranslator*>::iterator i = translations.find(name);
    if (i != translations.end())
        return i->second;
    else
        return 0;
}

bool
StringTranslator::getCreateTranslation(const string_t name, StringTranslator*& ptr)
{
    std::map<string_t,StringTranslator*>::iterator i = translations.find(name);
    if (i != translations.end()) {
        ptr = i->second;
        return true;
    }

    StringTranslator* rv = new StringTranslator();
    safeInsert(translations, name, rv);
    ptr = rv;
    return false;
}
